/*
    sysCheck v1.6.2

	Homebrew application which does several checks on installed IOS:
		- IOS Stub;
		- Fake signature (aka Trucha Bug);
		- ES_DiVerify (aka ES_Identify);
		- Flash Access;
		- NAND Access;
		- Boot2 Access;
		- USB 2.0.

	sysCheck generates a report on the root of the SD card (sd:/sysCheck.csv).




	This is my first C project so be kind!


	sysCheck is based on on the work of others. Credit to those guys:

		AnyTitle Deleter
			http://wiibrew.org/wiki/AnyTitle_Deleter

		DOP-Mii
			http://code.google.com/p/dop-mii

		IOS Check
			http://wiibrew.org/wiki/IOSCheck

		Libwiilight
			http://wiibrew.org/wiki/Libwiilight

		Preloader
			http://wiibrew.org/wiki/Preloader

		Priiloader
			http://code.google.com/p/priiloader

		Title Lister
			http://wiibrew.org/wiki/Homebrew_Apps/Title_Lister

		USB Loader GX
			http://code.google.com/p/usbloader-gui
*/

// Includes
#include <stdio.h>
#include <string.h>
#include <gccore.h>
#include <malloc.h>
#include <fat.h>
#include <unistd.h>
#include <wiilight.h>
#include <ogc/conf.h>

#include "..\include\sys.h"
#include "..\include\video.h"
#include "..\include\fatmounter.h"

// Filename
#define REPORT "sd:/sysCheck.csv"

// Auto detect language
#define AUTO_DETECT_LANGUAGE true

// Display homebrew title
void DisplayHomebrewTitle(bool isFrench)
{
	if (isFrench)
		PrintText("sysCheck v1.6.2 par Erik Spyder", false, false, true, BLUE, WHITE);
	else
		PrintText("sysCheck v1.6.2 by Erik Spyder", false, false, true, BLUE, WHITE);
}

// Main subroutine
int main(int argc, char **argv)
{
	// Detect the language settings:
	// 		CONF_LANG_JAPANESE
	// 		CONF_LANG_ENGLISH
	// 		CONF_LANG_GERMAN
	// 		CONF_LANG_FRENCH
	// 		CONF_LANG_SPANISH
	// 		CONF_LANG_ITALIAN
	// 		CONF_LANG_DUTCH
	// 		CONF_LANG_SIMP_CHINESE
	// 		CONF_LANG_TRAD_CHINESE
	// 		CONF_LANG_KOREAN
	bool isFrench;

	if (AUTO_DETECT_LANGUAGE)
		isFrench = (CONF_GetLanguage() == CONF_LANG_FRENCH);
	else
		isFrench = false;

	// Initialise the video
	InitialiseVideo();

	// Clear the console
	ClearConsole(WHITE);

	// Display the homebrew title
	DisplayHomebrewTitle(isFrench);

	// Get the console region
	if (isFrench)
		PrintText("\n  Obtention de la region de la console...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("\n  Getting the console region...\n", false, false, false, WHITE, BLACK);

    int regionSelection = CONF_GetRegion();

    // Get the system menu version
	if (isFrench)
		PrintText("  Obtention de la version du menu systeme...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Getting the system menu version...\n", false, false, false, WHITE, BLACK);

	u32 sysVersion = GetSysMenuVersion(isFrench);

    // Get the running IOS version and revision
	if (isFrench)
		PrintText("  Obtention du numero de version et de revision de l'IOS courant...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Getting the running IOS version and revision...\n", false, false, false, WHITE, BLACK);
	
	u32 runningIOS = IOS_GetVersion();
	u32 runningIOSRevision = IOS_GetRevision();

	// Get the console ID
	if (isFrench)
		PrintText("  Obtention de l'identifiant de la console et de la version du boot2...\n\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Getting the console ID and the boot2 version...\n\n", false, false, false, WHITE, BLACK);

	u32 deviceID = GetDeviceID(isFrench);

    // Get the boot2 version
	u32 boot2version = GetBoot2Version(isFrench);

	// Display the console region
	PrintText("  Region", false, false, false, WHITE, BLUE);

    switch (regionSelection)
	{
		case CONF_REGION_US:
			PrintText(": NTSC-U.", false, false, false, WHITE, BLUE);
			break;

		case CONF_REGION_EU:
			PrintText(": PAL.", false, false, false, WHITE, BLUE);
			break;

		case CONF_REGION_JP:
			PrintText(": NTSC-J.", false, false, false, WHITE, BLUE);
			break;

		case CONF_REGION_KR:
			PrintText(": KOR.", false, false, false, WHITE, BLUE);
			break;

		default:
			if (isFrench)
				PrintText(" inconnue!", false, false, false, WHITE, RED);
			else
				PrintText(" unknown!", false, false, false, WHITE, RED);

			break;
    }

	// Display the menu system version and the running IOS version and revision
	char textToDisplay[72];

	if (isFrench)
		sprintf(textToDisplay, "\n  Le menu systeme v%d s'execute sous l'IOS%d (rev %d).\n", sysVersion, runningIOS, runningIOSRevision);
	else
		sprintf(textToDisplay, "\n  The system menu v%d is running under IOS%d (rev %d).\n", sysVersion, runningIOS, runningIOSRevision);
	
	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	// Display Hollywood version
	sprintf(textToDisplay, "  Hollywood v0x%x.\n", *(u32 *)0x80003138);

	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	// Display the console ID
	if (isFrench)
		sprintf(textToDisplay, "  Identifiant de la console: %d.\n", deviceID);
	else
		sprintf(textToDisplay, "  Console ID: %d.\n", deviceID);

	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	// Display the boot2 version
	sprintf(textToDisplay, "  Boot2 v%u.\n\n", boot2version);
	
	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	// Get number of titles
	if (isFrench)
		PrintText("  Obtention du nombre de titres... ", false, false, false, WHITE, BLACK);
	else
		PrintText("  Getting number of titles... ", false, false, false, WHITE, BLACK);

	u32 nbTitles;

	if (ES_GetNumTitles(&nbTitles) < 0)
	{
		if (isFrench)
			PrintText("\n\n  Impossible d'obtenir le nombre de titres!\n", false, false, false, WHITE, RED);
		else
			PrintText("\n\n  Failed getting number of titles!\n", false, false, false, WHITE, RED);

		return false;
    }

	if (isFrench)
		sprintf(textToDisplay, "%d titres trouves.\n", nbTitles);
	else
		sprintf(textToDisplay, "Found %d titles.\n", nbTitles);

	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	// Allocate the memory for titles
	u64 *titles = memalign(32, nbTitles*sizeof(u64));

	if (titles == NULL)
	{
		if (isFrench)
			sprintf(textToDisplay, "  Impossible d'allouer de la memoire pour le titre %d!\n", nbTitles);
		else
			sprintf(textToDisplay, "  Unable to allocate the memory for %d titles!\n", nbTitles);

		PrintText(textToDisplay, false, false, false, WHITE, RED);

		return false;
    }

	// Get list of titles
	if (isFrench)
		PrintText("  Obtention de la liste des titres...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Getting list of titles...\n", false, false, false, WHITE, BLACK);

	if (ES_GetTitles(titles, nbTitles) < 0)
	{
		if (isFrench)
			PrintText("  Impossible d'obtenir la liste des titres!\n", false, false, false, WHITE, RED);
		else
			PrintText("  Failed getting list of titles!\n", false, false, false, WHITE, RED);

		return false;
    }

	int i;
	int countIOS = 0; // Number of IOS
	int countStubs = 0; // Number of IOS Stubs
	u32 titleID;

	s32 infoRevision[nbTitles];
	bool infoStub[nbTitles];

	// For each titles found
	for (i = 0; i < nbTitles; i++)
	{
		// Skip non-system titles
		if (titles[i] >> 32 != 1) continue;

		// Skip the system menu
		titleID = titles[i] & 0xFFFFFFFF;

		if (titleID == 2) continue;

		// Skip BC, MIOS and possible other non-IOS titles
		if (titleID > 0xFF) continue;

		// Skip the running IOS
		if (titleID == 0) continue;

		// Check if this title is an IOS stub
		u32 tmdSize;
		tmd *iosTMD ATTRIBUTE_ALIGN(32);

		// Get the stored TMD size for the title
		if (ES_GetStoredTMDSize(0x0000000100000000ULL | titleID, &tmdSize) < 0)
		{
			if (isFrench)
				sprintf(textToDisplay, "  Erreur lors de l'appel a ES_GetStoredTMDSize pour l'IOS%d!\n", titleID);
			else
				sprintf(textToDisplay, "  Failed to get the stored TMD size for IOS%d!\n", titleID);

			PrintText(textToDisplay, false, false, false, WHITE, RED);

			return false;
		}

		signed_blob *iosTMDBuffer = (signed_blob *)memalign(32, (tmdSize+32)&(~31));
		memset(iosTMDBuffer, 0, tmdSize);

		// Get the stored TMD for the title
		if (ES_GetStoredTMD(0x0000000100000000ULL | titleID, iosTMDBuffer, tmdSize) < 0)
		{
			if (isFrench)
				sprintf(textToDisplay, "  Erreur lors de l'appel a ES_GetStoredTMD pour l'IOS%d!\n", titleID);
			else
				sprintf(textToDisplay, "  Failed to get the stored TMD for IOS%d!\n", titleID);

			PrintText(textToDisplay, false, false, false, WHITE, RED);

			return false;
		}

		iosTMD = (tmd *)SIGNATURE_PAYLOAD(iosTMDBuffer);
		
		free(iosTMDBuffer);

		// Get the title version
		u8 noVersion = iosTMD->title_version;
		bool isStub = false;
		
		// Check if this is an IOS stub (according to WiiBrew.org)
		if (IsKnownStub(titleID, iosTMD->title_version))
			isStub = true;
		else
		{
			// If the version is 00, it's probably a stub
			//
			// Stubs have these things in common:
			//	- Title version is mostly 65280, or even better, the last 2 hexadecimal digits are 0;
			// 	- Stub have one app of their own (type 0x1) and 2 shared apps (type 0x8001).
			if (noVersion == 0)
				isStub = ((iosTMD->num_contents == 3) && (iosTMD->contents[0].type == 1 && iosTMD->contents[1].type == 0x8001 && iosTMD->contents[2].type == 0x8001));
			else		
				isStub = false;
		}

		// Add to IOS list
		titles[countIOS] = titles[i];
		infoRevision[countIOS] = iosTMD->title_version;
		infoStub[countIOS] = isStub;

		countIOS++;

		if (isStub) countStubs++;
	}

	if (isFrench)
		sprintf(textToDisplay, "\n  Il y a %d IOS et %d 'IOS Stubs' presents sur cette console.\n\n", (countIOS-countStubs), countStubs);
	else
		sprintf(textToDisplay, "\n  Found %d IOS and %d IOS Stubs on this console.\n\n", (countIOS-countStubs), countStubs);

	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	if (isFrench)
		PrintText("  Initialisation du 'FAT File System'...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Initialising the FAT file system...\n", false, false, false, WHITE, BLACK);

	// Initialise the FAT file system
	if (!fatInitDefault())
	{
		if (isFrench)
			PrintText("  Impossible d'initialiser le 'FAT File System'!\n", false, false, false, WHITE, RED);
		else
			PrintText("  Unable to initialise the FAT file system!\n", false, false, false, WHITE, RED);

        return false;
	}

    // Create the report
	FILE *file = fopen(REPORT, "w");

    if(!file)
	{
		if (isFrench)
			PrintText("  Impossible d'ouvrir le fichier!\n", false, false, false, WHITE, RED);
		else
			PrintText("  Unable to open the file!\n", false, false, false, WHITE, RED);

        return false;
    }

    // Create the report
	if (isFrench)
		fprintf(file, "sysCheck v1.6.2 par Erik Spyder\n\n");
	else
		fprintf(file, "sysCheck v1.6.2 by Erik Spyder\n\n");

    // Display the console region
	fprintf(file, "Region");

    switch (regionSelection)
	{
		case CONF_REGION_US:
			fprintf(file, ": NTSC-U");
			break;

		case CONF_REGION_EU:
			fprintf(file, ": PAL");
			break;

		case CONF_REGION_JP:
			fprintf(file, ": NTSC-J");
			break;

		case CONF_REGION_KR:
			fprintf(file, ": KOR");
			break;

		default:
			if (isFrench)
				fprintf(file, " inconnue");
			else
				fprintf(file, " unknown");

			break;
    }

	if (isFrench)
	{
		fprintf(file, "\nLe menu systeme v%d s'execute sous l'IOS%d (rev %d)\n", sysVersion, runningIOS, IOS_GetRevision());
		fprintf(file, "Hollywood v0x%x\n", *(u32 *)0x80003138);
		fprintf(file, "Identifiant de la console: %d\n", deviceID);
		fprintf(file, "Boot2 v%u\n", boot2version);
		fprintf(file, "%d titres trouves\n", nbTitles);

		if ((countIOS-countStubs) == 0)
			fprintf(file, "Il n'y a aucun IOS sur cette console\n");
		else
			fprintf(file, "%d IOS presents sur cette console\n", (countIOS-countStubs));

		if (countStubs == 0)
			fprintf(file, "Il n'y a aucun 'IOS Stub' sur cette console\n");
		else
			fprintf(file, "%d 'IOS Stubs' presents sur cette console\n", countStubs);
    }
	else
	{
		fprintf(file, "\nThe system menu v%d is running under IOS%d (rev %d)\n", sysVersion, runningIOS, IOS_GetRevision());
		fprintf(file, "Hollywood v0x%x\n", *(u32 *)0x80003138);
		fprintf(file, "Console ID: %d\n", deviceID);
		fprintf(file, "Boot2 v%u\n", boot2version);
		fprintf(file, "Found %d titles\n", nbTitles);

		if ((countIOS-countStubs) == 0)
			fprintf(file, "No IOS on this console\n");
		else
			fprintf(file, "Found %d IOS on this console\n", (countIOS-countStubs));

		if (countStubs == 0)
			fprintf(file, "No IOS Stub on this console\n");
		else
			fprintf(file, "Found %d IOS Stubs on this console\n", countStubs);
    }

	titles = realloc(titles, countIOS*sizeof(u64));
	nbTitles = countIOS;

	if (nbTitles == 0)
	{
		fclose(file);

		return false;
	}

	// Close the report
	fclose(file);

	// Sort IOS titles
	if (isFrench)
		PrintText("  Tri des titres...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Sorting titles...\n", false, false, false, WHITE, BLACK);

	u64 lowestTitle;
	s32 lowestRevision;
	bool lowestStub;
	int lowestID;
	int j;

	for (i = 0; i < nbTitles; i++)
	{
		lowestTitle = titles[i];
		lowestRevision = infoRevision[i];
		lowestStub = infoStub[i];
		lowestID = i;
		
		for (j = i+1; j < nbTitles; j++)
		{
			if (titles[j] < lowestTitle)
			{
				lowestTitle = titles[j];
				lowestRevision = infoRevision[j];
				lowestStub = infoStub[j];
				lowestID = j;
			}
		}

		// Swap IOS title
		titles[lowestID] = titles[i];
		titles[i] = lowestTitle;

		// Swap IOS revision
		infoRevision[lowestID] = infoRevision[i];
		infoRevision[i] = lowestRevision;

		// Swap IOS stub
		infoStub[lowestID] = infoStub[i];
		infoStub[i] = lowestStub;
	}

	// Get the certificates from the NAND
	if (isFrench)
		PrintText("  Obtention des certificats de la 'NAND'... ", false, false, false, WHITE, BLACK);
	else
		PrintText("  Getting the certificates from the NAND... ", false, false, false, WHITE, BLACK);

	if (GetCertificates(isFrench))
	{
		if (isFrench)
			PrintText("Succes!\n", false, false, false, WHITE, BLUE);
		else
			PrintText("Success!\n", false, false, false, WHITE, BLUE);
	}
	else
	{
		fclose(file);

        return false;
	}

	// Test vulnerabilities in IOS
	if (isFrench)
		PrintText("  Test de vulnerabilite de chacun des IOS... Un moment SVP...\n\n  ", false, false, false, WHITE, BLACK);
	else
		PrintText("  Testing IOS vulnerabilities... Please wait...\n\n  ", false, false, false, WHITE, BLACK);

	bool infoFakeSignature[nbTitles];
	bool infoESIdentify[nbTitles];
	bool infoFlashAccess[nbTitles];
	bool infoNANDAccess[nbTitles];
	bool infoBoot2Access[nbTitles];
	bool infoUSB2[nbTitles];

	for (i = 0; i < nbTitles; i++)
	{
		PrintText(" ", false, false, false, GREEN, GREEN);

		if (infoStub[i])
		{
			infoFakeSignature[i] = false;
			infoESIdentify[i] = false;
			infoFlashAccess[i] = false;
			infoNANDAccess[i] = false;
			infoBoot2Access[i] = false;
			infoUSB2[i] = false;
		}
		else
		{
			// Get the title ID
			titleID = titles[i] & 0xFFFFFFFF;

			// Reload IOS
			IOS_ReloadIOS(titleID);	

			// Test fake signature
			infoFakeSignature[i] = (CheckFakeSignature());

			// Test ES Identify
			infoESIdentify[i] = (CheckESIdentify());

			// Test Flash Access
			infoFlashAccess[i] = (CheckFlashAccess());

			// Test NAND Access
			infoNANDAccess[i] = (CheckNANDAccess());

			// Test Boot2 Access
			infoBoot2Access[i] = (CheckBoot2Access());

			// Test USB 2.0
			infoUSB2[i] = (CheckUSB2(titleID));
		}
	}

	// Initialise the Wii Light
	WIILIGHT_Init();

	// Wait 1 second
	sleep(1);

	// Clear the console
	ClearConsole(WHITE);

	// Display the homebrew title
	DisplayHomebrewTitle(isFrench);

	// Turn on the Wii Light
	WiiLightControl(WIILIGHT_ON);

	if (isFrench)
		sprintf(textToDisplay, "\n  Rechargement de l'IOS%d (rev %d)...\n", runningIOS, runningIOSRevision);
	else
		sprintf(textToDisplay, "\n  Reloading IOS%d (rev %d)...\n", runningIOS, runningIOSRevision);

	PrintText(textToDisplay, false, false, false, WHITE, BLACK);

	// Wait 2 seconds
	sleep(2);

	// Reload the running IOS
	IOS_ReloadIOS(runningIOS);	

	if (isFrench)
		PrintText("  Montage de la carte SD...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Mounting the SD Card...\n", false, false, false, WHITE, BLACK);

	// Mount the SD Card
	MountSD();

	if (isFrench)
		PrintText("  Initialisation du 'FAT File System'...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Initialising the FAT file system...\n", false, false, false, WHITE, BLACK);

	// Initialise the FAT file system
	if (!fatInitDefault())
	{
		if (isFrench)
			PrintText("  Impossible d'initialiser le 'FAT File System'!\n", false, false, false, WHITE, RED);
		else
			PrintText("  Unable to initialise the FAT file system!\n", false, false, false, WHITE, RED);

        return false;
	}

	if (isFrench)
		PrintText("  Generation du rapport...\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Generating the report...\n", false, false, false, WHITE, BLACK);

	// Generate the CSV report
	file = fopen(REPORT, "a");

    if(!file)
	{
		if (isFrench)
			PrintText("  Impossible d'ouvrir le fichier!\n", false, false, false, WHITE, RED);
		else
			PrintText("  Unable to open the file!\n", false, false, false, WHITE, RED);

        return false;
	}
	else
	{
		// Display the header
		if (isFrench)
			fprintf(file, "\nIOS (revision);IOS Stub;Trucha Bug;ES Identify;Flash Access;NAND Access;Boot2 Access;USB 2.0\n");
		else
			fprintf(file, "\nIOS (revision),IOS Stub,Trucha Bug,ES Identify,Flash Access,NAND Access,Boot2 Access,USB 2.0\n");

		// Display IOS vulnerabilities
		for (i = 0; i < nbTitles; i++)
		{
			titleID = titles[i] & 0xFFFFFFFF;

			if (isFrench)
			{
				fprintf(file, "IOS%d (rev %d);", titleID, infoRevision[i]);

				// Check BootMii As IOS (BootMii As IOS is installed on IOS254 rev 31338)
				if (titleID == 254 && infoRevision[i] == 31338)
					fprintf(file, "Non;?;?;?;?;?;?");
				else
				{
					if (infoStub[i])
						fprintf(file, "Oui;?;?;?;?;?;?");
					else
					{
						fprintf(file, "Non;");
						fprintf(file, "%sctive;", (infoFakeSignature[i]) ? "A" : "Desa");
						fprintf(file, "%sctive;", (infoESIdentify[i])    ? "A" : "Desa");
						fprintf(file, "%sctive;", (infoFlashAccess[i])   ? "A" : "Desa");
						fprintf(file, "%sctive;", (infoNANDAccess[i])    ? "A" : "Desa");
						fprintf(file, "%sctive;", (infoBoot2Access[i])   ? "A" : "Desa");
						fprintf(file, "%sctive",  (infoUSB2[i])          ? "A" : "Desa");
					}
				}
			}
			else
			{
				fprintf(file, "IOS%d (rev %d),", titleID, infoRevision[i]);

				if (titleID == 254 && infoRevision[i] == 31338)
					fprintf(file, "No,?,?,?,?,?,?");
				else
				{
					if (infoStub[i])
						fprintf(file, "Yes,?,?,?,?,?,?");
					else
					{
						fprintf(file, "No,");
						fprintf(file, "%sabled,", (infoFakeSignature[i]) ? "En" : "Dis");
						fprintf(file, "%sabled,", (infoESIdentify[i])    ? "En" : "Dis");
						fprintf(file, "%sabled,", (infoFlashAccess[i])   ? "En" : "Dis");
						fprintf(file, "%sabled,", (infoNANDAccess[i])    ? "En" : "Dis");
						fprintf(file, "%sabled,", (infoBoot2Access[i])   ? "En" : "Dis");
						fprintf(file, "%sabled",  (infoUSB2[i])          ? "En" : "Dis");
					}
				}
			}

			fprintf(file, "\n");
		}
	}

    // Get and display the current date and time
	struct tm today;
	time_t rawtime;

	time(&rawtime);
	today = *localtime(&rawtime);

	if (isFrench)
		fprintf(file, "\n\nRapport genere le %4.4d/%2.2d/%2.2d", today.tm_year + 1900, today.tm_mon + 1, today.tm_mday);
	else
		fprintf(file, "\n\nReport generated on %4.4d/%2.2d/%2.2d", today.tm_year + 1900, today.tm_mon + 1, today.tm_mday);

	// Close the report
	fclose(file);

	if (isFrench)
		PrintText("  Demontage de la carte SD...\n\n", false, false, false, WHITE, BLACK);
	else
		PrintText("  Unmounting the SD Card...\n\n", false, false, false, WHITE, BLACK);

	// Unmount the SD Card
	UnmountSD();

	if (isFrench)
		sprintf(textToDisplay, "  %s genere avec succes.", REPORT);
	else
		sprintf(textToDisplay, "  %s generated successfully.", REPORT);

	PrintText(textToDisplay, false, false, false, WHITE, BLUE);

	if (isFrench)
		PrintText("\n\n\n  Retour au 'Loader' dans 5 secondes...", false, false, false, WHITE, BLACK);
	else
		PrintText("\n\n\n  Returning to the loader in 5 seconds...", false, false, false, WHITE, BLACK);
	
	// Wait 2 seconds
	sleep(2);

	// Turn off the Wii Light
	WiiLightControl(WIILIGHT_OFF);

	// Wait 3 seconds
	sleep(3);

	// Return to System Menu
	//SYS_ResetSystem(SYS_RETURNTOMENU, 0, 0);

	// Reboot Wii
	//SYS_ResetSystem(SYS_RESTART, 0, 0);

	// Return to the loader
	return true;
}
